import os

import click

from inoft_vocal_framework.cli.cli_cache import CliCache
from inoft_vocal_framework.databases.dynamodb.dynamodb import DynamoDbMessagesAdapter
from inoft_vocal_framework.speechs.ssml_builder_core import SpeechsList

""""
@click.command()
@click.argument("push")
def hello(push):
    click.echo("Hello, %s!" % push)

@click.command()
@click.argument('src', nargs=-1)
@click.argument('dst', nargs=1)
def copy(src, dst):
    for fn in src:
        click.echo('move %s to folder %s' % (fn, dst))
"""

@click.group()
def cli():
    pass

@cli.command("new")
def new():
    from inoft_vocal_framework.cli import new as new_module
    new_module.new()

@cli.command("deploy")
def deploy():
    from inoft_vocal_framework.cli.deploy import DeployHandler as deployHandler
    deployHandler().handle()

@cli.group()
def messages():
    pass

@messages.command()
@click.option('--file', prompt="Filepath of file containing the message to push", type=str,
              default=CliCache.cache().get("lastMessagesFilepath").to_str())
def push(file):
    CliCache.cache().put("lastMessagesFilepath", file)
    nice_quit()

    click.echo('Pushing the messages')
    if os.path.exists(file):
        import importlib
        import importlib.util
        spec = importlib.util.spec_from_file_location("messages", file)
        messages_module_file = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(messages_module_file)

        messages_db = DynamoDbMessagesAdapter(is_admin_mode=True, table_name="test_messages", region_name="eu-west-3")

        messages_file_vars_dict = vars(messages_module_file)
        for var_key, var_object in messages_file_vars_dict.items():
            if isinstance(var_object, SpeechsList):
                messages_db.post_new_speechs_list(speechs_list=var_object)
                print(f"Posted speech category : {var_key}")
    else:
        if click.confirm("File did not exist, do you want to select a new one ?"):
            filepath = click.prompt("Filepath of the file")
            push(file=filepath)

    nice_quit()

def nice_quit():
    CliCache.save_cache_to_yaml()

if __name__ == '__main__':
    new()




class Response:
    json_key: str

    def __init__(self):
        self._text = str()

    def say(self, text_or_ssml: str):
        self.text = text_or_ssml

    def reprompt(self, text_or_ssml: str):
        print("Reprompt of bixby not yet implemented")

    @property
    def text(self) -> str:
        return self._text

    @text.setter
    def text(self, text: str) -> None:
        if not isinstance(text, str):
            raise Exception(f"text was type {type(text)} which is not valid value for his parameter.")
        self._text = text

    def to_dict(self):
        return {
            "text": self.text
        }




class DummyObject:
    # When we try to access any attribute of the
    # dummy object, it will return himself.
    def __getattr__(self, item):
        return self

    # And if dummy object is called, nothing is being done,
    # and error created if the function did not exist.
    def __call__(self, *args, **kwargs):
        pass



import os
from pathlib import Path
from typing import Tuple, Optional, Any

import click
from cerberus import Validator
from inoft_vocal_framework.safe_dict import SafeDict


class Settings:
    _settings = None
    settings_loaded = False
    last_settings_filepath = None
    last_settings_file_extension_type = None

    class ExtendedValidator(Validator):
        def _validate_check_database_field_present(self, check_database_field_present: bool, field: str, value):
            """ {'type': 'boolean'} """
            if check_database_field_present is True and value not in self.document.keys():
                self._error(field, f"The key {value} was missing from the following section : {self.document}")

        def _validate_is_database_not_disabled(self, is_database_not_disabled: bool, field: str, value):
            """ {'type': 'boolean'} """
            if is_database_not_disabled is True and "disable_database" in self.document.keys() and self.document["disable_database"]:
                self._error(field, f"The database is disabled in the following section : {self.document}")

    @property
    def _settings_file_validator_schema(self) -> dict:
        return {
            "default_session_data_timeout": {
                "required": True,
                "type": "integer"
            },
            "sessions_users_data": {
                "required": True,
                "type": "dict",
                "schema": {
                    "disable_database": {
                        "required": False,
                        "type": "boolean",
                    },
                    "database_client": {
                        "required": True,
                        "type": "string",
                        "allowed": ["dynamodb"],
                        "is_database_not_disabled": True,
                        "check_database_field_present": True,
                    },
                    "dynamodb": {
                        "type": "dict",
                        "schema": {
                            "table_name": {
                                "required": True,
                                "type": "string"
                            },
                            "region_name": {
                                "required": True,
                                "type": "string"
                            }
                        }
                    }
                }
            },
            "messages": {
                "required": True,
                "type": "dict",
                "schema": {
                    "use_database_dynamic_messages": {
                        "required": True,
                        "type": "boolean"
                    },
                    "database_client": {
                        "type": "string",
                        "allowed": ["dynamodb"],
                        "required": True,
                        "check_database_field_present": True,
                        "dependencies": {"use_database_dynamic_messages": True}
                    },
                    "dynamodb": {
                        "type": "dict",
                        "schema": {
                            "table_name": {
                                "required": True,
                                "type": "string"
                            },
                            "region_name": {
                                "required": True,
                                "type": "string"
                            }
                        }
                    }
                }
            },
            "deployment": {
                "type": "dict",
                "schema": {
                    "api_gateway_id": {
                        "type": "string",
                    },
                    "s3_bucket_name": {
                        "type": "string",
                    },
                    "lambda_name": {
                        "type": "string",
                    },
                    "endpoints": {
                        "type": "dict",
                        "schema": {
                            "alexaApiEndpointUrlNotRecommendedToUse": {
                                "type": "string"
                            },
                            "googleAssistantApiEndointUrl": {
                                "type": "string"
                            },
                            "samsungBixbyApiEndointUrl": {
                                "type": "string"
                            },
                            "siriApiEndointUrl": {
                                "type": "string"
                            },
                        }
                    }
                }
            }
        }



    @property
    def settings(self) -> SafeDict:
        if Settings.settings_loaded is not True:
            raise Exception(f"The settings have not yet been loaded and are : {Settings.settings}")
        return Settings._settings

    @settings.setter
    def settings(self, settings_dict: dict):
        validator = self.ExtendedValidator()
        is_valid = validator.validate(settings_dict, self._settings_file_validator_schema)
        if is_valid is not True:
            raise Exception(f"The settings file was not valid. Please modify it or recreate it : {validator.errors}")
        else:
            Settings._settings = SafeDict(settings_dict)
            Settings.settings_loaded = True

    def find_load_settings_file(self, root_folderpath: str):
        def list_settings_filepaths(dirpath: str) -> list:
            found_settings_filepaths = []
            for root_dirpath, dirs, filenames in os.walk(dirpath):
                if "\\inoft_vocal_framework\\templates" not in root_dirpath:
                    # The app_settings file of the templates of the framework
                    # are not considered as valid app_settings file.
                    for filename in filenames:
                        if Path(filename).stem == "app_settings":
                            found_settings_filepaths.append(os.path.join(root_dirpath, filename))
            return found_settings_filepaths

        found_files = list_settings_filepaths(root_folderpath)

        file_path_object = None
        while file_path_object is None:
            print(f"Found {len(found_files)} settings file at following paths :")
            for i_file, file in enumerate(found_files):
                print(f"{i_file + 1} - {file}")
            file_index = str(click.prompt("What is the number of the file you wish to use ?"))
            if not file_index.isdigit() or int(file_index) > len(found_files) or int(file_index) < 0:
                print("Please write a valid file index in the form of an int like 1 or 3")
            else:
                selected_filepath = found_files[int(file_index) - 1]
                print(f"Selected settings filepath : {click.style(text=selected_filepath, fg='blue')}")
                file_path_object = Path(selected_filepath)


        if file_path_object.suffix == ".yaml":
            self.load_yaml(file_path_object)
        elif file_path_object.suffix == ".json":
            self.load_json(file_path_object)
        else:
            raise Exception(f"The only supported settings file extension are .yaml and"
                            f" .json but your file was in {file_path_object.suffix}")

    def load_yaml(self, settings_file: str):
        from yaml import safe_load, YAMLError
        with open(settings_file, "r") as file_stream:
            try:
                self.settings = safe_load(file_stream)
            except YAMLError as error:
                raise Exception(f"The yaml file was not valid, and caused an error when loading."
                                f"Please check the file or recreate it with the cli : {error}")
        Settings.last_settings_filepath = settings_file
        Settings.last_settings_file_extension_type = "yaml"

    def load_json(self, settings_file: str):
        from json import load as json_load
        with open(settings_file, "r") as file_stream:
            try:
                self.settings = json_load(file_stream)
            except Exception as error:
                raise Exception(f"The json file was not valid, and caused an error when loading."
                                f"Please check the file or recreate it with the cli : {error}")
        Settings.last_settings_filepath = settings_file
        Settings.last_settings_file_extension_type = "json"

    def save_settings(self):
        from json import dumps as json_dumps
        from yaml import safe_dump as yaml_dump

        if Settings.last_settings_file_extension_type in ["yaml", "json"]:
            settings_dict = self.settings.to_dict()

            with open(Settings.last_settings_filepath, "w+") as file_stream:
                settings_dict = self.settings.to_dict()
                file_stream.write(yaml_dump(settings_dict) if Settings.last_settings_file_extension_type == "yaml" else json_dumps(settings_dict))
        else:
            raise Exception("Please load a valid yaml or json settings file before attempting to save them.")




from inoft_vocal_framework.safe_dict import SafeDict


class Request:
    LaunchKeyName = "launch"

    def __init__(self):
        self._userId = str()
        self._sessionId = str()
        self._intent = str()
        self._parameters = SafeDict()

    def is_launch_request(self):
        if self.intent == self.LaunchKeyName:
            return True
        else:
            return False

    def is_in_intent_names(self, intent_names_list):
        if isinstance(intent_names_list, list):
            if self.intent in intent_names_list:
                return True
        elif isinstance(intent_names_list, str):
            if self.intent == intent_names_list:
                return True
        return False

    def get_intent_parameter_value(self, parameter_key: str, default=None):
        return self.parameters.get(dict_key=parameter_key).to_any(default=default)

    @property
    def userId(self) -> str:
        return self._userId

    @userId.setter
    def userId(self, userId: str) -> None:
        if not isinstance(userId, str):
            raise Exception(f"userId was type {type(userId)} which is not valid value for his parameter.")
        self._userId = userId

    @property
    def sessionId(self) -> str:
        return self._sessionId

    @sessionId.setter
    def sessionId(self, sessionId: str) -> None:
        if not isinstance(sessionId, str):
            raise Exception(f"sessionId was type {type(sessionId)} which is not valid value for his parameter.")
        self._sessionId = sessionId

    @property
    def intent(self) -> str:
        return self._intent

    @intent.setter
    def intent(self, intent: str) -> None:
        if not isinstance(intent, str):
            raise Exception(f"intent was type {type(intent)} which is not valid value for his parameter.")
        self._intent = intent

    @property
    def parameters(self) -> SafeDict:
        if isinstance(self._parameters, dict):
            self._parameters = SafeDict(self._parameters)
        return self._parameters

    @parameters.setter
    def parameters(self, parameters: dict) -> None:
        if not isinstance(parameters, dict):
            raise Exception(f"parameters was type {type(parameters)} which is not valid value for his parameter.")
        self._parameters = SafeDict(parameters)





from collections import Callable

from inoft_vocal_framework.exceptions import raise_if_variable_not_expected_type, raise_if_value_not_in_list
from inoft_vocal_framework.platforms_handlers.nested_object_to_dict import NestedObjectToDict
from inoft_vocal_framework.utils.general import is_text_ssml


class Image:
    json_key = "image"

    def __init__(self, small_image_url: str, large_image_url: str):
        self._smallImageUrl = small_image_url
        self._largeImageUrl = large_image_url

    @property
    def smallImageUrl(self) -> str:
        return self._smallImageUrl

    @smallImageUrl.setter
    def smallImageUrl(self, smallImageUrl: str) -> None:
        if not isinstance(smallImageUrl, str):
            raise Exception(f"smallImageUrl was type {type(smallImageUrl)} which is not valid value for his parameter.")
        self._smallImageUrl = smallImageUrl

    @property
    def largeImageUrl(self) -> str:
        return self._largeImageUrl

    @largeImageUrl.setter
    def largeImageUrl(self, largeImageUrl: str) -> None:
        if not isinstance(largeImageUrl, str):
            raise Exception(f"largeImageUrl was type {type(largeImageUrl)} which is not valid value for his parameter.")
        self._largeImageUrl = largeImageUrl

class Card:
    json_key = "card"

    type_simple = "Simple"
    type_standard = "Standard"
    type_link_account = "LinkAccount"
    type_ask_for_permissions_content = "AskForPermissionsConsent"
    available_types = [type_simple, type_standard, type_link_account, type_ask_for_permissions_content]

    def __init__(self, type_value: str, title: str, text: str = None, content_text: str = None, image: Image = None):
        self._type = type_value
        self._title = title
        self._text = text
        self._content = content_text
        self._image = image

    def do_not_include(self):
        if (self._type is None) or (self._title == "" and self._content == ""):
            return True
        else:
            return False

    @property
    def type(self):
        return self._type

    @type.setter
    def type(self, type_value: str) -> None:
        if type_value not in self.available_types:
            raise Exception(f"The type_value {type_value} is not a supported type in {self.available_types}")
        self._type = type_value

    @property
    def title(self):
        return self._title

    @title.setter
    def title(self, title: str) -> None:
        self._title = title

    @property
    def text(self) -> str:
        return self._text

    @text.setter
    def text(self, text: str) -> None:
        if not isinstance(text, str):
            raise Exception(f"text was type {type(text)} which is not valid value for his parameter.")
        self._text = text

    @property
    def content(self):
        return self._content

    @content.setter
    def content(self, content: str) -> None:
        self._content = content

    @property
    def image(self) -> Image:
        if self._image is None:
            self._image = Image()
        return self._image

class Directives(list):
    json_key = "directives"

    def __init__(self):
        super().__init__()

    @property
    def audioPlayer(self):
        from inoft_vocal_framework.platforms_handlers.alexa_v1.audioplayer.audioplayer_directives import AudioPlayer
        for directive in self:
            if isinstance(directive, AudioPlayer):
                return directive

    def add_audio_player(self, played_type: str, play_behavior: str, token_identifier: str, url: str, offsetInMilliseconds: int = 0):
        if self.audioPlayer is not None:
            raise Exception(f"AudioPlayer has already been set and cannot be set twice")

        from inoft_vocal_framework.platforms_handlers.alexa_v1.audioplayer.audioplayer_directives import AudioPlayer
        self.append(AudioPlayer(played_type=played_type, play_behavior=play_behavior,
                                token_identifier=token_identifier, url=url,
                                offsetInMilliseconds=offsetInMilliseconds))

    def do_not_include(self) -> bool:
        return True if len(self) == 0 else False


class OutputSpeech:
    TYPE_KEY_TEXT = "PlainText"
    TYPE_KEY_SSML = "SSML"
    json_key = "outputSpeech"

    def __init__(self):
        self._type = None
        self._text = None
        self._ssml = None

    def is_speech_empty(self):
        if self._type == self.TYPE_KEY_TEXT:
            return True if self.text is None else False
        elif self._type == self.TYPE_KEY_SSML:
            return True if self.ssml is None else False
        else:
            return True

    @property
    def text(self):
        return self._text

    @text.setter
    def text(self, text: str) -> None:
        raise_if_variable_not_expected_type(value=text, expected_type=str, variable_name="text")
        self._type = "PlainText"
        self._text = text

    @property
    def ssml(self):
        return self._ssml

    @ssml.setter
    def ssml(self, ssml: str) -> None:
        raise_if_variable_not_expected_type(value=ssml, expected_type=str, variable_name="ssml")
        self._type = "SSML"
        self._ssml = ssml

    def set_based_on_type(self, value_to_set: str, type_key: str):
        if type_key == self.TYPE_KEY_SSML:
            self.ssml = value_to_set
        elif type_key == self.TYPE_KEY_TEXT:
            self.text = value_to_set
        else:
            raise Exception(f"Type key {type_key} is not supported.")
        return self

    def do_not_include(self):
        return self.is_speech_empty()

class Reprompt:
    json_key = "reprompt"

    def __init__(self):
        self.outputSpeech = OutputSpeech()

    def do_not_include(self):
        return self.outputSpeech.is_speech_empty()

class Response:
    json_key = "response"

    def __init__(self):
        self._outputSpeech = OutputSpeech()
        self._reprompt = Reprompt()
        self._directives = Directives()
        self._card = None
        self._shouldEndSession = False

    def say(self, text_or_ssml: str):
        if is_text_ssml(text_or_ssml=text_or_ssml) is True:
            self.outputSpeech.ssml = text_or_ssml
        else:
            self.outputSpeech.text = text_or_ssml

    def say_reprompt(self, text_or_ssml: str):
        if is_text_ssml(text_or_ssml=text_or_ssml) is True:
            self.reprompt.outputSpeech.ssml = text_or_ssml
        else:
            self.reprompt.outputSpeech.text = text_or_ssml

    @property
    def audioplayer(self):
        from inoft_vocal_framework.platforms_handlers.alexa_v1.audioplayer.audioplayer_directives import AudioPlayer
        for directive in self.directives:
            if isinstance(directive, AudioPlayer):
                return directive

    # todo: to remove its deprecated (use the audioplayer object)
    def play_audio(self, identifier: str,  mp3_file_url: str,
                   title: str, subtitle: str, icon_image_url: str, background_image_url: str,
                   milliseconds_start_offset: int = 0, played_type: str = None, play_behavior: str = None,
                   override_default_end_session: bool = False):

        from inoft_vocal_framework.platforms_handlers.alexa_v1.audioplayer.audioplayer_directives import AudioPlayer
        if played_type is None:
            played_type = AudioPlayer.TYPE_PLAY
        if play_behavior is None:
            play_behavior = AudioPlayer.PLAY_BEHAVIOR_REPLACE_ALL

        # todo: check mp3 file validity
        self.directives.add_audio_player(played_type=played_type, play_behavior=play_behavior,
                                         token_identifier=identifier, url=mp3_file_url,
                                         offsetInMilliseconds=milliseconds_start_offset)

        if override_default_end_session is False:
            self.end_session()
        # When playing an audio file, by default we end the session so that the file will be played (it can be overridden with an argument)

    def end_session(self, should_end: bool = True):
        self.shouldEndSession = should_end

    @property
    def outputSpeech(self) -> OutputSpeech:
        return self._outputSpeech

    @property
    def reprompt(self) -> Reprompt:
        return self._reprompt

    @property
    def directives(self) -> Directives:
        return self._directives

    @property
    def card(self):
        return self._card

    @card.setter
    def card(self, card: Card) -> None:
        raise_if_variable_not_expected_type(value=card, expected_type=Card, variable_name="card")
        self._card = card

    @property
    def shouldEndSession(self):
        return self._shouldEndSession

    @shouldEndSession.setter
    def shouldEndSession(self, shouldEndSession: bool) -> None:
        raise_if_variable_not_expected_type(value=shouldEndSession, expected_type=bool, variable_name="shouldEndSession")
        self._shouldEndSession = shouldEndSession

    def to_dict(self) -> dict:
        dict_object = NestedObjectToDict.get_dict_from_nested_object(object_to_process=self,
                                                                     key_names_identifier_objects_to_go_into=["json_key"])
        dict_object["version"] = "1.0"
        dict_object["sessionAttributes"] = dict()
        return dict_object

if __name__ == "__main__":
    NestedObjectToDict.get_dict_from_nested_object(Response(), ["json_key"])



from inoft_vocal_framework.exceptions import raise_if_variable_not_expected_type, raise_if_value_not_in_list


class Context:
    json_key = "context"

    class AudioPlayer:
        json_key = "AudioPlayer"
        PLAYER_ACTIVITY_STOPPED = "STOPPED"
        AVAILABLE_PLAYER_ACTIVITIES = [PLAYER_ACTIVITY_STOPPED]

        def __init__(self):
            self._offsetInMilliseconds = None
            self._token = None
            self._playerActivity = None

        @property
        def offsetInMilliseconds(self) -> int:
            return self._offsetInMilliseconds

        @offsetInMilliseconds.setter
        def offsetInMilliseconds(self, offsetInMilliseconds: int) -> None:
            raise_if_variable_not_expected_type(value=offsetInMilliseconds, expected_type=int, variable_name="offsetInMilliseconds")
            self._offsetInMilliseconds = offsetInMilliseconds

        @property
        def token(self) -> str:
            return self._token

        @token.setter
        def token(self, token: str) -> None:
            raise_if_variable_not_expected_type(value=token, expected_type=str, variable_name="token")
            self._token = token

        @property
        def playerActivity(self) -> str:
            return self._playerActivity

        @playerActivity.setter
        def playerActivity(self, playerActivity: str) -> None:
            raise_if_value_not_in_list(value=playerActivity, list_object=self.AVAILABLE_PLAYER_ACTIVITIES, variable_name="playerActivity")
            self._playerActivity = playerActivity

    class System:
        json_key = "System"

        class Application:
            json_key = "application"

            def __init__(self):
                self._applicationId = None

            @property
            def applicationId(self) -> str:
                return self._applicationId

            @applicationId.setter
            def applicationId(self, applicationId: str) -> None:
                raise_if_variable_not_expected_type(value=applicationId, expected_type=str, variable_name="applicationId")
                self._applicationId = applicationId

        class User:
            json_key = "user"

            def __init__(self):
                self._userId = None

            @property
            def userId(self) -> str:
                return self._userId

            @userId.setter
            def userId(self, userId: str) -> None:
                raise_if_variable_not_expected_type(value=userId, expected_type=str, variable_name="userId")
                self._userId = userId

        class Device:
            json_key = "device"

            class SupportedInterfaces:
                json_key = "supportedInterfaces"

                def __init__(self):
                    pass
                    # todo: complete the supported interfaces class

            def __init__(self):
                self._deviceId = None
                self.supportedInterfaces = self.SupportedInterfaces()

            @property
            def deviceId(self) -> str:
                return self._deviceId

            @deviceId.setter
            def deviceId(self, deviceId: str) -> None:
                raise_if_variable_not_expected_type(value=deviceId, expected_type=str, variable_name="deviceId")
                self._deviceId = deviceId

        def __init__(self):
            self.application = self.Application()
            self.user = self.User()
            self.device = self.Device()
            self._apiEndpoint = None
            self._apiAccessToken = None

        @property
        def apiEndpoint(self) -> str:
            return self._apiEndpoint

        @apiEndpoint.setter
        def apiEndpoint(self, apiEndpoint: str) -> None:
            raise_if_variable_not_expected_type(value=apiEndpoint, expected_type=str, variable_name="apiEndpoint")
            self._apiEndpoint = apiEndpoint

        @property
        def apiAccessToken(self) -> str:
            return self._apiAccessToken

        @apiAccessToken.setter
        def apiAccessToken(self, apiAccessToken: str) -> None:
            raise_if_variable_not_expected_type(value=apiAccessToken, expected_type=str, variable_name="apiAccessToken")
            self._apiAccessToken = apiAccessToken

    def __init__(self):
        self.audioPlayer = self.AudioPlayer()
        self.system = self.System()



def get_function_or_class_from_file_and_path(file_filepath: str, path_qualname: str):
    from os.path import isfile

    if not isfile(file_filepath):
        print(f"SERIOUS WARNING ! The module file containing a callback function at filepath  {file_filepath} "
              f"has not been found. The callback cannot be used. If your interaction that use the callback is not "
              f"working, it is because you have issues with the location of your file containing the callback.")
    else:
        import importlib
        import importlib.util
        module_spec = importlib.util.spec_from_file_location("fileContainingFunctionOrClass", file_filepath)
        if module_spec is not None:
            module_file = importlib.util.module_from_spec(module_spec)
            module_spec.loader.exec_module(module_file)

            path_list = path_qualname.split(".")
            if len(path_list) > 0:
                vars_module_file = vars(module_file)
                if path_list[0] in vars_module_file:
                    if len(path_list) == 1:
                        # If the path list is only of length one, and that the first element is
                        # found in the vars of the module, we can return it right away.
                        return vars_module_file[path_list[0]]
                    else:
                        # Otherwise we will start a recursive function to handle potentially nested classes or functions.
                        from inspect import getmembers

                        def get_nested_class_or_function(current_nested_class_or_function, path_remaining_elements: list):
                            if len(path_remaining_elements) > 0:
                                class_or_function_name = path_remaining_elements[0]
                                if class_or_function_name == "<locals>":
                                    raise Exception(
                                        f"A callback function cannot be a nested function of another function ({class_or_function_name}) "
                                        "Please make it available from the root of the file or from a class then relaunch the event "
                                        "that inserted the function path of the callback function (redo the entire interaction up to"
                                        "the point where you set the callback to an event).")

                                members_nested_class_or_function = getmembers(current_nested_class_or_function)
                                for tuple_member in members_nested_class_or_function:
                                    if tuple_member[0] == class_or_function_name:
                                        if len(path_remaining_elements) > 1:
                                            return get_nested_class_or_function(current_nested_class_or_function=tuple_member[1],
                                                path_remaining_elements=path_remaining_elements[1:])
                                        else:
                                            return tuple_member[1]
                            return None

                        return get_nested_class_or_function(current_nested_class_or_function=vars_module_file[path_list[0]],
                                                            path_remaining_elements=path_list[1:])
    return None



from json import dumps as json_dumps
from inoft_vocal_framework.platforms_handlers.handler_input import HandlerInput


class LambdaResponseWrapper:
    def __init__(self, response_dict: dict):
        if not isinstance(response_dict, dict):
            raise Exception(f"The response_dict must be of type dict and is of type : {type(response_dict)}")

        self._response_dict = response_dict

    def get_wrapped(self, handler_input: HandlerInput) -> dict:
        if handler_input.is_alexa_v1 is True:
            return self._response_dict

        elif (handler_input.is_dialogflow_v1 is True
        or handler_input.is_bixby_v1 is True):
            return {
                "isBase64Encoded": False,
                "statusCode": 200,
                "headers": dict(),
                "body": json_dumps(self._response_dict)
            }

        else:
            raise Exception(f"Platform is not supported.")



class SafeDict:
    def __init__(self, classic_dict=None, overloaded_navigated_dict=None):
        if classic_dict is None:
            classic_dict = dict()
        if isinstance(classic_dict, list):
            output_dict = dict()
            for i in range(len(classic_dict)):
                output_dict[i] = classic_dict[i]
            classic_dict = output_dict
        elif not isinstance(classic_dict, dict):
            try:
                classic_dict = dict(classic_dict)
            except Exception:
                try:
                    classic_dict = dict(vars(classic_dict))
                except Exception:
                    print(f"The following variable could not be converted to a dict in order to create a SafeDict object : {classic_dict}")

        if isinstance(classic_dict, dict):
            self.classic_dict = classic_dict
            if not isinstance(overloaded_navigated_dict, dict):
                self.navigated_dict = self.classic_dict
            else:
                self.navigated_dict = overloaded_navigated_dict
        else:
            self.classic_dict = dict()
            if not isinstance(overloaded_navigated_dict, dict):
                self.navigated_dict = dict()
            else:
                self.navigated_dict = overloaded_navigated_dict

    
    def reset_navigated_dict(self):
        self.navigated_dict = self.classic_dict

    def retrieve_navigated_dict_object(self, reset_navigated_dict=True):
        if isinstance(self.navigated_dict, dict) or isinstance(self.navigated_dict, list):
            # A dict or a list will not make a copy of the value, but
            # point a reference to the object, so we need to copy them.
            navigated_dict_object = self.navigated_dict.copy()
        else:
            navigated_dict_object = self.navigated_dict

        if reset_navigated_dict is True:
            self.reset_navigated_dict()
        return navigated_dict_object

    def retrieve_navigated_dict_values(self, reset_navigated_dict=True):
        if isinstance(self.navigated_dict, dict):
            navigated_dict_values = list(self.navigated_dict.values())

            if len(navigated_dict_values) > 1:
                navigated_dict_values = navigated_dict_values
            elif len(navigated_dict_values) > 0:
                navigated_dict_values = navigated_dict_values[0]
            else:
                navigated_dict_values = None

            if reset_navigated_dict is True:
                self.reset_navigated_dict()
            return navigated_dict_values
        else:
            navigated_dict_values = self.navigated_dict
            self.reset_navigated_dict()
            return navigated_dict_values

    def get(self, dict_key: str):
        if isinstance(self.navigated_dict, dict) and dict_key in self.navigated_dict.keys():
            self.navigated_dict = self.navigated_dict[dict_key]
        else:
            self.navigated_dict = None
        return self

    def get_set(self, dict_key: str, value_to_set_if_missing={}):
        if isinstance(self.navigated_dict, dict):
            if dict_key not in self.navigated_dict.keys():
                self.navigated_dict[dict_key] = value_to_set_if_missing

            self.navigated_dict = self.navigated_dict[dict_key]
        return self

    def put(self, dict_key, value_to_put):
        if isinstance(self.navigated_dict, dict) and dict_key is not None:
            if value_to_put is None:
                value_to_put = ""
            self.navigated_dict[dict_key] = value_to_put
        else:
            print(f"Warning ! Tried to put a value into a dict of a SafeDict, but the current navigated dict was not pointing to a dict object !")
        return self

    def pop(self, dict_key: str):
        if isinstance(self.navigated_dict, dict) and isinstance(dict_key, str) and dict_key in self.navigated_dict.keys():
            self.navigated_dict.pop(dict_key)
        else:
            print(f"Warning ! Tried to pop a key of a dict of a SafeDict, but the current navigated dict was not pointing to a dict object !")
        return self

    def process_any_safedicts_in_dict_or_list_to_dict(self, dict_or_list_to_process):
        if isinstance(dict_or_list_to_process, list):
            for i in range(len(dict_or_list_to_process)):
                dict_or_list_to_process[i] = self.process_any_safedicts_in_dict_or_list_to_dict(dict_or_list_to_process[i])
            return dict_or_list_to_process

        elif isinstance(dict_or_list_to_process, dict):
            for key in dict_or_list_to_process.keys():
                dict_or_list_to_process[key] = self.process_any_safedicts_in_dict_or_list_to_dict(dict_or_list_to_process[key])
            return dict_or_list_to_process

        elif isinstance(dict_or_list_to_process, SafeDict):
            return dict_or_list_to_process.to_dict()

        else:
            return dict_or_list_to_process

    def to_str(self, default="", reset_navigated_dict=True) -> str:
        navigated_dict_values = self.retrieve_navigated_dict_values(reset_navigated_dict=reset_navigated_dict)
        if isinstance(navigated_dict_values, str):
            return navigated_dict_values
        else:
            return default

    def to_bool(self, default=False, reset_navigated_dict=True) -> bool:
        navigated_dict_values = self.retrieve_navigated_dict_values(reset_navigated_dict=reset_navigated_dict)
        if isinstance(navigated_dict_values, bool):
            return navigated_dict_values
        else:
            return default

    def to_int(self, default=0, reset_navigated_dict=True) -> int:
        navigated_dict_values = self.retrieve_navigated_dict_values(reset_navigated_dict=reset_navigated_dict)
        try:
            int_value = int(navigated_dict_values)
            return int_value
        except Exception as cannot_convert_error:
            return default
        
    def to_float(self, default=0.0, reset_navigated_dict=True) -> float:
        navigated_dict_values = self.retrieve_navigated_dict_values(reset_navigated_dict=reset_navigated_dict)
        try:
            float_value = float(navigated_dict_values)
            return float_value
        except Exception as cannot_convert_error:
            return default

    def to_list(self, default=[], reset_navigated_dict=True) -> list:
        navigated_dict_values = self.retrieve_navigated_dict_values(reset_navigated_dict=reset_navigated_dict)
        navigated_dict_values = self.process_any_safedicts_in_dict_or_list_to_dict(dict_or_list_to_process=navigated_dict_values)
        if isinstance(navigated_dict_values, list):
            return navigated_dict_values
        elif isinstance(navigated_dict_values, dict):
            return list(navigated_dict_values.values())
        else:
            return default

    def to_dict(self, default={}, reset_navigated_dict=True) -> dict:
        navigated_dict_object = self.retrieve_navigated_dict_object(reset_navigated_dict=reset_navigated_dict)
        navigated_dict_object = self.process_any_safedicts_in_dict_or_list_to_dict(dict_or_list_to_process=navigated_dict_object)
        if isinstance(navigated_dict_object, dict):
            return navigated_dict_object
        elif isinstance(navigated_dict_object, list):
            output_dict = dict()
            for i in range(len(navigated_dict_object)):
                output_dict[i] = navigated_dict_object[i]
            return output_dict
        else:
            return default

    def to_safedict(self, default=None, reset_navigated_dict=True):
        navigated_dict_object = self.retrieve_navigated_dict_object(reset_navigated_dict=reset_navigated_dict)
        if navigated_dict_object is not None:
            return SafeDict(navigated_dict_object)
        else:
            return default if default is not None else SafeDict()

    def to_any(self, default=None, reset_navigated_dict=True):
        navigated_dict_object = self.retrieve_navigated_dict_object(reset_navigated_dict=reset_navigated_dict)
        return navigated_dict_object

    def to_specific_type(self, type_to_return: type, reset_navigated_dict=True):
        if type_to_return == str:
            return self.to_str(reset_navigated_dict=reset_navigated_dict)
        elif type_to_return == bool:
            return self.to_bool(reset_navigated_dict=reset_navigated_dict)
        elif type_to_return == int:
            return self.to_int(reset_navigated_dict=reset_navigated_dict)
        elif type_to_return == float:
            return self.to_float(reset_navigated_dict=reset_navigated_dict)
        elif type_to_return == list:
            return self.to_list(reset_navigated_dict=reset_navigated_dict)
        elif type_to_return == dict:
            return self.to_dict(reset_navigated_dict=reset_navigated_dict)
        elif type_to_return == SafeDict:
            return self.to_safedict(reset_navigated_dict=reset_navigated_dict)
        else:
            raise Exception(f"The following type is not supported by the SafeDict object : {type_to_return}."
                            f"If you post a request in the issues section of the GitHub,"
                            f"i will add the object type that you need to the SafeDict class - Robinson")

    def keys(self):
        return self.navigated_dict.keys()

    def copy(self, reset_navigated_dict=True):
        new_safeDict_copy = SafeDict(classic_dict=self.classic_dict, overloaded_navigated_dict=self.navigated_dict)
        if reset_navigated_dict:
            self.reset_navigated_dict()
        return new_safeDict_copy



import os
import json

import yaml

from inoft_vocal_framework.safe_dict import SafeDict


class CliCache:
    _cache = None

    @staticmethod
    def get_cli_cache_filepath() -> str:
        filepath = os.path.join(os.path.dirname(os.path.abspath(__file__)), "cli_cache.yaml")
        if not os.path.isfile(filepath):
            with open(filepath, "w+") as cache_file:
                cache_file.write(yaml.safe_dump({}))
        return filepath

    @staticmethod
    def cache() -> SafeDict:
        if CliCache._cache is None:
            CliCache.load_cache_from_yaml()
        return CliCache._cache

    @staticmethod
    def load_cache_from_json():
        expected_cli_cache_filepath = CliCache.get_cli_cache_filepath()
        if os.path.exists(expected_cli_cache_filepath):
            with open(expected_cli_cache_filepath) as cache_file:
                CliCache._cache = SafeDict(json.load(cache_file))

    @staticmethod
    def load_cache_from_yaml():
        expected_cli_cache_filepath = CliCache.get_cli_cache_filepath()
        if os.path.exists(expected_cli_cache_filepath):
            with open(expected_cli_cache_filepath) as cache_file:
                CliCache._cache = SafeDict(yaml.safe_load(cache_file))

    @staticmethod
    def save_cache_to_json():
        with open(CliCache.get_cli_cache_filepath(), "w+") as cache_file:
            cache_file.write(json.dumps(CliCache.cache().to_dict()))

    @staticmethod
    def save_cache_to_yaml():
        with open(CliCache.get_cli_cache_filepath(), "w+") as cache_file:
            cache_file.write(yaml.safe_dump(CliCache.cache().to_dict()))
